"""This module manages Bluetooth connection.

BluetoothManager allows finding Bittle's physical address, connecting to
Bittle, sending and receiving messages from it.
"""

import socket
import subprocess

import bluetooth
import serial.tools.list_ports


__author__ = "EnriqueMoran"


class BluetoothManager:
    """Main class to manage Bluetooth connection.

    Attributes
    ----------
    name : str
        Bittle device name (by default its BittleSPP-XXXXXX).
    address : str
        Bittle device MAC address.
    port : int
        Communication port.
    discovery_timeout : int
        Time for discovery Bluetooth devices (seconds).
    recv_timeout : int
        Socket timeout for receiving messages (seconds).
    socket : bluetooth.BluetoothSocket
        Socket for Bluetooth connection.

    Methods
    -------
    initialize_name_and_address(get_first_bittle=True):
        Finds and sets Bittle's device name and MAC address.
    get_paired_devices(flush_cache=True, lookup_names=True):
        Returns avaliable paired devices.
    connect():
        Connects to Bittle.
    send_msg(msg):
        Sends a message to Bittle.
    recv_msg(buffer_size=1024):
        Returns received message from Bittle.
    close_connection():
        Closes connection with Bittle.
    """

    def __init__(self):
        self._name = ""
        self._address = ""
        self._port = 1
        self._discovery_timeout = 8
        self._recv_timeout = 10
        self.socket = bluetooth.BluetoothSocket(bluetooth.RFCOMM)

    def __del__(self):
        self.socket.close()

    def __repr__(self):
        return f"BluetoothManager - name: {self.name}, address: " \
               f"{self.address}, port: {self.port}, discovery_timeout: " \
               f"{self.discovery_timeout}, recv_timeout: {self.recv_timeout}"

    def __str__(self):
        return f"BluetoothManager - name: {self.name}, address: " \
               f"{self.address}, port: {self.port}"

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, new_name):
        if isinstance(new_name, str) and new_name:
            self._name = new_name
        else:
            raise TypeError("Name must be non empty str.")

    @property
    def address(self):
        return self._address

    @address.setter
    def address(self, new_address):
        if isinstance(new_address, str) and new_address:
            self._address = new_address
        else:
            raise TypeError("Address must be non empty str.")

    @property
    def port(self):
        return self._port

    @port.setter
    def port(self, new_port):
        if isinstance(new_port, int) and new_port > 0:
            self._port = new_port
        else:
            raise TypeError("Port type must be int, greater than 0.")

    @property
    def discovery_timeout(self):
        return self._discovery_timeout

    @discovery_timeout.setter
    def discovery_timeout(self, new_timeout):
        if isinstance(new_timeout, int) and new_timeout > 0:
            self._discovery_timeout = new_timeout
        else:
            raise TypeError("New timeout type must be int, greater than 0.")

    @property
    def recv_timeout(self):
        return self._recv_timeout

    @recv_timeout.setter
    def recv_timeout(self, new_timeout):
        if isinstance(new_timeout, int) and new_timeout > 0:
            self._recv_timeout = new_timeout
        else:
            raise TypeError("New timeout type must be int, greater than 0.")

    def initialize_name_and_address(self, get_first_bittle=True):
        """Sets self._name and self._address values by searching
        among paired devices and returns its values.

        Parameters:
            get_first_bittle (bool): If True, it will search for the first
            "BittleSPP" occurrence; otherwise will search for self._name
            ocurrence (full name must be stored in self._name:
            BittleSPP-XXXXXX). If is set to False but there is no valid
            self._name (empty), it will work as if was set to True.

        Returns:
            name (str) : Found name, None if not found.
            address (str) : Found address, None if not found.
        """
        if isinstance(get_first_bittle, bool):
            pass
        else:
            raise TypeError("Value type must be bool.")

        search_name = self.name if not get_first_bittle and self.name else \
            "BittleSPP"
        paired_devices = self.get_paired_devices()

        for address, name in list(paired_devices):
            if search_name in name:
                self.name = name
                self.address = address
                return self.name, self.address
        else:  # Bittle name not found, return (None, None, None)
            return None, None

    def get_paired_devices(self):
        """Returns A list of (MAC address, device name) tuples with paired
        devices.

        Check bluetooth.discover_devices documentation for more info.
        """
        return bluetooth.discover_devices(duration=self.discovery_timeout,
                                          flush_cache=True,
                                          lookup_names=True)

    def connect(self):
        """Connects to Bittle.

        Connects to Bittle and wait until full response is given
        (response will contain "Finished! at the end").
        Once its connected, set self.socket's timeout to self._recv_timeout.

        Returns:
            res (bool) : True if connected succesfully, False otherwise.
        """
        res = False
        try:
            self.socket.connect((self.address, self.port))
            self.socket.settimeout(self._recv_timeout)
            while True:
                data = self.socket.recv(1024)  # TODO: adjust buffer size
                if len(data) == 0:
                    break
                elif b"Finished!" in data:
                    res = True
                    break
        except:
            pass
        if not res:  # Reset socket
            self.socket = bluetooth.BluetoothSocket(bluetooth.RFCOMM)
        return res

    def send_msg(self, msg):
        """Sends a message to Bittle.

        Parameters:
            msg (str) : Message to send.
        """
        if isinstance(msg, str) and msg:
            self.socket.send(msg)
        else:
            raise TypeError("Message must be non empty str.")

    def recv_msg(self, buffer_size=1024):
        """Receives a message from Bittle.

        Parameters:
            buffer_size (int) : Buffer size.

        Returns:
            data (bytes) : Received data.
        """
        data = b''
        if isinstance(buffer_size, int) and buffer_size > 0:
            try:
                data = self.socket.recv(buffer_size)
            except socket.error as err:
                raise socket.error("{!s}".format(err)) from None
        else:
            raise TypeError("Buffer size must be int, greater than zero.")
        return data

    def close_connection(self):
        """Closes connection.
        """
        self.socket.close()



"""This module manages Serial communication.

SerialManager allows sending commands to Bittle through Serial.
"""

import serial
import serial.tools.list_ports

__author__ = "EnriqueMoran"


class SerialManager:
    """Main class to manage Serial connection.

    Attributes
    ----------
    port : str
        Serial communication port.
    baudrate : int
        Baud rate.
    timeout : int
        Serial communication timeout (seconds).
    parity : int
        Serial communication parity (possible values: none, odd, even).
    serial : serial.Serial
        Serial communication instance.

    Methods
    -------
    initialize():
        Sets serial communication parameters. If any of the parameters is
        updated after initialization, this method must be called to apply
        the changes.
    discover_port():
        Searches among avaliable communication ports the one associated
        to CH340 USB driver, which is used by Bittle.
    connect():
        Starts serial communication. Return wether connection was achieved.
    close_connection():
        Closes serial communication.
    send_msg(msg):
        Sends a message to Bittle.
    recv_msg():
        Returns received message from Bittle (byte).
    """

    def __init__(self):
        self._port = "COM1"
        self._baudrate = 115200
        self._timeout = 5
        self._parity = serial.PARITY_NONE
        self.serial = serial.Serial()

    def __del__(self):
        self.serial.close()

    def __repr__(self):
        return f"SerialManager - port: {self.port}, baudrate: " \
               f"{self.baudrate}, timeout: {self.timeout}, parity: " \
               f"{self.parity}"

    def __str__(self):
        return f"SerialManager - port: {self.port}, baudrate: " \
               f"{self.baudrate}, timeout: {self.timeout}"

    @property
    def port(self):
        return self._port

    @port.setter
    def port(self, new_port):
        if isinstance(new_port, str) and new_port:
            self._port = new_port
        else:
            raise TypeError("Port must be non empty str.")

    @property
    def baudrate(self):
        return self._baudrate

    @baudrate.setter
    def baudrate(self, new_baudrate):
        if isinstance(new_baudrate, int) and new_baudrate > 0:
            self._baudrate = new_baudrate
        else:
            raise TypeError("Baudrate type must be int, greater than 0.")

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, new_timeout):
        if isinstance(new_timeout, int) and new_timeout >= 0:
            self._timeout = new_timeout
        else:
            raise TypeError("Timeout must be positive int.")

    @property
    def parity(self):
        return self._parity

    @parity.setter
    def parity(self, new_parity):
        if isinstance(new_parity, str) and new_port:
            if new_parity.lower().replace(' ', '') == 'none':
                self._parity = serial.PARITY_NONE
            elif new_parity.lower().replace(' ', '') == 'odd':
                self._parity = serial.PARITY_ODD
            elif new_parity.lower().replace(' ', '') == 'even':
                self._parity = serial.PARITY_EVEN
            else:
                raise TypeError("Parity value be 'none', 'odd' or 'even'.")
        else:
            raise TypeError("Parity must be non empty str.")

    def initialize(self):
        """Sets serial communication parameters.
        """
        self.serial.baudrate = self.baudrate
        self.serial.port = self.port
        self.serial.timeout = self.timeout
        self.serial.parity = self.parity

    def discover_port(self):
        """Search among avaliable communication ports the one associated
        to CH340 USB driver, which is used by Bittle; if found, set self.port
        with its value. Returns True if found, False otherwise.
        """
        res = False
        for port in serial.tools.list_ports.comports():
            if "CH340" in port.description:
                self.port = port.device
                res = True
                break
        return res

    def connect(self):
        """Connects to Bittle and wait until full response is given
        (response will contain "Finished! at the end").
        Returns:
            res (bool) : True if connected successfully, False otherwise.
        """
        res = False
        self.serial.open()
        while True:
            data = self.recv_msg()
            if len(data) == 0:
                    break
            elif b"Finished!" in data:
                res = True
                self.recv_msg()  # Remove last blank line
                break
        return res


    def close_connection(self):
        """Closes serial communication.
        """
        self.serial.close()

    def send_msg(self, msg):
        """Sends a message to Bittle.
        """
        if isinstance(msg, str) and msg:
            self.serial.write(msg.encode())
        else:
            raise TypeError("Message must be non empty str.")

    def recv_msg(self):
        """Reads a serial data line (till '\n' character).

        Returns:
            data (byte) : Received data.
        """
        return self.serial.readline()



"""Connect to Bittle, control and manage its behaviour.

bittleManager is a high level module that allows connecting to Bittle through
Bluetooth or WiFi, define Bittle's behaviour and send/receive commands to
control it.
"""

import uuid

from enum import Enum

from pyBittle.bluetoothManager import *
from pyBittle.serialManager import *
from pyBittle.wifiManager import *

__author__ = "EnriqueMoran"


class Command(Enum):
    """Defines avaliable commands.
    """
    REST = 1
    FORWARD = 2
    GYRO = 3
    LEFT = 4
    BALANCE = 5
    RIGHT = 6
    SHUTDOWN = 7
    BACKWARD = 8
    CALIBRATION = 9
    STEP = 10
    CRAWL = 11
    WALK = 12
    TROT = 13
    LOOKUP = 14
    BUTTUP = 15
    RUN = 16
    BOUND = 17
    GREETING = 18
    PUSHUP = 19
    PEE = 20
    STRETCH = 21
    SIT = 22
    ZERO = 23
    BUNNY = 24
    BACKFLIP = 25
    SLEEP = 26
    CHECKAROUND = 27


class Gait(str, Enum):
    """Defines avaliable gaits.
    """
    WALK = 'kwk'
    CRAWL = 'kcr'
    TROT = 'ktr'
    RUN = 'krn'


class Direction(str, Enum):
    """Defines avaliable movement directions.
    """
    FORWARD = 'F'
    FORWARDLEFT = 'L'
    FORWARDRIGHT = 'R'
    BACKWARD = 'B'
    BACKWARDLEFT = 'BL'
    BACKWARDRIGHT = 'BR'


class Bittle:
    """High level class that represents your Bittle.

    Attributes
    ----------
    id : uuid.UUID
        Bittle's unique id.
    bluetoothManager : BluetoothManager
        Manager for sending messages to Bittle through Bluetooth connection.
    wifiManager : WifiManager
        Manager for sending messages to Bittle through WiFi connection.
    serialManager : SerialManager
        Manager for sending messages to Bittle through Serial connection.
    gait : Gait
        Current gait.
    commands : {Command: str}
        Avaliable commands that can be sent to Bittle.

    Methods
    -------
    connect_bluetooth(get_first_bittle):
        Connects to Bittle through Bluetooth connection.
    send_command_bluetooth(command):
        Sends a command to Bittle through Bluetooth connection.
    send_msg_bluetooth(message):
        Sends a custom message to Bittle through Bluetooth connection.
    receive_msg_bluetooth(buffer_size):
        Returns received message from Bittle through Bluetooth connection.
    send_movement_bluetooth(direction):
        Sends a movement command to Bittle through Bluetooth connection.
    disconnect_bluetooth():
        Closes Bluetooth connection with Bittle.
    has_wifi_connection():
        Checks wether there is connection with REST API.
    send_command_wifi(command):
        Sends a command to Bittle through WiFi connection.
    send_msg_wifi(message):
        Sends a custom message to Bittle through WiFi connection.
    send_movement_wifi(direction):
        Sends a movement command to Bittle through WiFi connection.
    connect_serial(discover_port):
        Connects to Bittle through Serial connection.
    send_command_serial(command):
        Sends a command to Bittle through Serial connection.
    send_msg_serial(message):
        Sends a custom message to Bittle through Serial connection.
    receive_msg_serial(buffer_size):
        Returns received message from Bittle through Serial connection.
    send_movement_serial(direction):
        Sends a movement command to Bittle through Serial connection.
    disconnect_serial():
        Closes Serial connection with Bittle.
    """

    def __init__(self):
        self._id = uuid.uuid4()  # Bittle's id
        self.bluetoothManager = BluetoothManager()
        self.wifiManager = WifiManager()
        self.serialManager = SerialManager()
        self._gait = Gait.WALK  # Current gait
        self._commands = {  # Command : message to Bittle
            Command.REST: 'd',
            Command.FORWARD: 'F',
            Command.GYRO: 'g',
            Command.LEFT: 'L',
            Command.BALANCE: 'kbalance',
            Command.RIGHT: 'R',
            Command.SHUTDOWN: 'z',
            Command.BACKWARD: 'B',
            Command.CALIBRATION: 'c',
            Command.STEP: 'kvt',
            Command.CRAWL: 'kcr',
            Command.WALK: 'kwk',
            Command.TROT: 'ktr',
            Command.LOOKUP: 'klu',
            Command.BUTTUP: 'kbuttUp',
            Command.RUN: 'krn',
            Command.BOUND: 'kbd',
            Command.GREETING: 'khi',
            Command.PUSHUP: 'kpu',
            Command.PEE: 'kpee',
            Command.STRETCH: 'kstr',
            Command.SIT: 'ksit',
            Command.ZERO: 'kzero',
            Command.BUNNY: 'kbdF',
            Command.BACKFLIP: 'kbf',
            Command.SLEEP: 'kstp',
            Command.CHECKAROUND: 'kck'
        }

    def __eq__(self, other):
        return self._id == other._id

    def __str__(self):  # TODO: Complete
        return f"Bittle with id '{self._id}' Bluetooth name: " \
                f"'{self.bluetoothManager.name} ' MAC address: " \
                f"'{self.bluetoothManager.address}' " \
                f"IP address: '{self.wifiManager.ip}' " \
                f"Serial port: {self.serialManager.port}"

    @property
    def gait(self):
        return self._gait

    @gait.setter
    def gait(self, new_gait):
        if isinstance(new_gait, Gait):
            self._gait = new_gait
        else:
            raise TypeError("New gait must be Gait type.")

    def connect_bluetooth(self, get_first_bittle=True):
        """Connects to Bittle.

        Parameters:
            get_first_bittle (bool): If True, connects to the first
            "BittleSPP" found device, otherwise connects to
            bluetoothManager.name device.

        Returns:
            res (bool) : True if connected, False otherwise.
        """
        res = False
        name, addr = self.bluetoothManager.initialize_name_and_address(
                     get_first_bittle)
        if name and addr:  # Bittle found among avaliable paired devices
            res = self.bluetoothManager.connect()
        return res

    def send_command_bluetooth(self, command):
        """Sends command to Bittle through Bluetooth connection.

        Parameters:
            command (Comand) : Command to send.
        """
        if isinstance(command, Command):
            self.bluetoothManager.send_msg(self._commands[command])
        else:
            raise TypeError("Command type must be Command.")

    def send_msg_bluetooth(self, message):
        """Sends custom message to Bittle through Bluetooth connection.

        Parameters:
            message (str) : Message to send.
        """
        if isinstance(message, str):
            self.bluetoothManager.send_msg(message)
        else:
            raise TypeError("Message type must be str.")

    def receive_msg_bluetooth(self, buffer_size=1024):
        """Receives a message from Bittle through Bluetooth connection.

        Parameters:
            buffer_size (int) : Buffer size.

        Returns:
            data (bytes) : Received data.
        """
        return self.bluetoothManager.recv_msg(buffer_size)

    def send_movement_bluetooth(self, direction):
        """Sends movement commands with current gait through Bluetooth
        connection.
        """
        if isinstance(direction, Direction):
            command = ''
            if direction in [Direction.BACKWARD, Direction.BACKWARDLEFT,
                             Direction.BACKWARDRIGHT]:
                command = 'kbk' + direction[1:]
            else:
                command = self.gait.value + direction.value
            self.send_msg_bluetooth(command)
        else:
            raise TypeError("Direction must be Direction type.")

    def disconnect_bluetooth(self):
        """Closes Bluetooth connection.
        """
        self.bluetoothManager.close_connection()

    def has_wifi_connection(self):
        """Returns True if there is connection with REST API, False otherwise.
        """
        return self.wifiManager.has_connection()

    def send_command_wifi(self, command):
        """Sends command to Bittle through WiFi connection.

        Parameters:
            command (Comand) : Command to send.

        Returns:
            res (int) : REST API response code, -1 if
            there is no connection.
        """
        if isinstance(command, Command):
            return self.wifiManager.send_msg(self._commands[command])
        else:
            raise TypeError("Command type must be Command.")

    def send_msg_wifi(self, message):
        """Sends custom message to Bittle through WiFi connection.

        Parameters:
            message (str) : Message to send.

        Returns:
            res (int) : REST API response code, -1 if
            there is no connection.
        """
        if isinstance(message, str):
            return self.wifiManager.send_msg(message)
        else:
            raise TypeError("Message type must be str.")

    def send_movement_wifi(self, direction):
        """Sends movement commands with current gait through WiFi
        connection.
        """
        if isinstance(direction, Direction):
            command = ''
            if direction in [Direction.BACKWARD, Direction.BACKWARDLEFT,
                             Direction.BACKWARDRIGHT]:
                command = 'kbk' + direction[1:]
            else:
                command = self.gait.value + direction.value
            self.send_msg_wifi(command)
        else:
            raise TypeError("Direction must be Direction type.")

    def connect_serial(self, discover_port=True):
        """Connects to Bittle.

        Parameters:
            discover_port (bool): If True, connects to the first
            communication port associated to CH340, otherwise connects to
            serialManager.port.

        Returns:
            res (bool) : True if connected, False otherwise.
        """
        if isinstance(discover_port, bool):
            pass
        else:
            raise TypeError("Value type must be bool.")
        res = False
        if discover_port:
            port_found = self.serialManager.discover_port()
        self.serialManager.initialize()
        res = self.serialManager.connect()
        return res

    def send_command_serial(self, command):
        """Sends command to Bittle through serial connection.

        Parameters:
            command (Comand) : Command to send.
        """
        if isinstance(command, Command):
            self.serialManager.send_msg(self._commands[command])
        else:
            raise TypeError("Command type must be Command.")

    def send_msg_serial(self, message):
        """Sends custom message to Bittle through serial connection.

        Parameters:
            message (str) : Message to send.
        """
        if isinstance(message, str):
            self.serialManager.send_msg(message)
        else:
            raise TypeError("Message type must be str.")

    def receive_msg_serial(self):
        """Receives a message from Bittle through serial connection.

        Returns:
            data (bytes) : Received data.
        """
        return self.serialManager.recv_msg()

    def send_movement_serial(self, direction):
        """Sends movement commands with current gait through serial
        connection.
        """
        if isinstance(direction, Direction):
            command = ''
            if direction in [Direction.BACKWARD, Direction.BACKWARDLEFT,
                             Direction.BACKWARDRIGHT]:
                command = 'kbk' + direction[1:]
            else:
                command = self.gait.value + direction.value
            self.send_msg_serial(command)
        else:
            raise TypeError("Direction must be Direction type.")

    def disconnect_serial(self):
        """Closes Serial connection.
        """
        self.serialManager.close_connection()


"""This module manages WiFi connection.

WifiManager allows sending commands to Bittle by using the ESP8266 REST API.
"""

import ipaddress
import requests


__author__ = "EnriqueMoran"


class WifiManager:
    """Main class to manage WiFi connection.

    Attributes
    ----------
    ip : str
        Bittle's ip address.
    http_address : str
        Bittle's REST API address.

    Methods
    -------
    get_status_code():
        Returns REST API actionpage request response code.
    has_connection():
        Returns True if there is connection to REST API, False otherwise.
    send_msg(msg):
        Sends a message to Bittle.
    """

    def __init__(self):
        self._ip = ""
        self._http_address = f""

    def __repr__(self):
        return f"WifiManager - ip: {self.ip}, " \
               f"http_address: {self.http_address}"

    @property
    def ip(self):
        return self._ip

    @ip.setter
    def ip(self, new_ip):
        if isinstance(new_ip, str) and new_ip:
            try:
                ipaddress.ip_address(new_ip)
            except:
                raise TypeError("Invalid IPv4 address.")
            self._ip = new_ip
            self._http_address = f"http://{new_ip}/"
        else:
            raise TypeError("IP must be non empty str.")

    @property
    def http_address(self):
        return self._http_address

    def get_status_code(self):
        """Returns Action Page request response.

        Returns:
            res (int) : Action Page request response code, -1 if
            there is no connection.
        """
        res = -1
        http_address = self.http_address + "actionpage"
        try:
            response = requests.get(http_address)
            res = response.status_code
        except:
            pass
        return res

    def has_connection(self):
        """Returns True if Action Page request response is 200, False
        otherwise.

        Returns:
            res (bool) : True if there is connection with REST API,
            False otherwise.
        """
        res = False
        http_address = self.http_address + "actionpage"
        try:
            response = requests.get(http_address)
            if response.status_code == 200:
                res = True
        except:
            pass
        return res

    def send_msg(self, msg):
        """Sends a message to Bittle. Returns request response (int).

        Parameters:
            msg (str) : Message to send.

        Returns:
            status_code (int) : Request response code, -1 if there is no
            connection with REST API.
        """
        res = -1
        if isinstance(msg, str) and msg:
            query = {'name': msg}
            http_address = self.http_address + "action"
            try:
                response = requests.get(http_address, params=query)
                res = response.status_code
            except:
                pass
            return res
        else:
            raise TypeError("Message must be non empty str.")



import os
import pathlib
import sys

from setuptools import find_packages, setup

VERSION = "1.1.3"
DIR = pathlib.Path(__file__).parent
README = (DIR / "README.md").read_text()

if sys.version_info[:2] < (3, 6):
    raise RuntimeError("Python version >= 3.6 required.")

setup(
    name='pyBittle',
    version=VERSION,
    description='Open source library for connecting and controlling Bittle.',
    long_description=README,
    long_description_content_type="text/markdown",
    url='https://github.com/EnriqueMoran/pyBittle',
    author='EnriqueMoran',
    author_email='enriquemoran95@gmail.com',
    install_requires=['pybluez', 'pyserial', 'requests'],
    packages=find_packages(),
    zip_safe=False,
)



from pyBittle.bittleManager import *
from pyBittle.bluetoothManager import *
from pyBittle.serialManager import *
from pyBittle.wifiManager import *

__author__ = "EnriqueMoran"

__version__ = "1.1.3"